/*
 * Andreea Dieaconu, 335CA
 */

#include <stdio.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>


#define sign(a) ( ( (a) < 0 )  ?  -1   : ( (a) > 0 ) )

#define ENC_OVF 0xFF 	/* timer counter overflow value */

#define ANGLE_ERROR_THRESHOLD 10 	/* angle error */
#define DIST_ERROR_THRESHOLD 10 	/* distance error */

#define pi 3.14
#define R 1.6 /* dimension of the wheel in cm */
#define N 250 /* number of ticks per revolution (one completed rotation of the wheel) */

typedef enum {
	MOTOR_A = 1, /* right motor */
	MOTOR_B = 2, /* letf motor */
} Motor;


typedef enum {
	BACKWARD = -1,
	FLOAT = 0,
	FORWARD = 1,
	BREAK = 2
} Direction;


/*
 * Structure for counting the ticks, for timer.
 * Count from 0 to 0xFF.
 */
typedef struct Counter {
	unsigned char old;
	unsigned char curr;

} Counter;


/*
 * Structure for counting the ticks, for robot.
 */
typedef struct Encoder {
	short old;
	short curr;

} Encoder;

/*
 * The position of the robot
 */
typedef struct Position{
  long x; /* measured in cm */
  long y; /* measured in cm */

  long teta; /* measured in radians from ox axe */

}Position;







/* Robot coords */
Position *robotPosition;

/* Goal's coords */
Position goal;

/* Counters for left and right wheel */
Counter *cl, *cr;

/* Encoders for left and right wheel */
Encoder *el, *er;

/* Distance traveled by left and right wheel */
short suml = 0, sumr = 0;

/* vars used for updating onces at 10ms */
short PWM1 = 0, PWM2 = 0;
char update = 0;



/*
 * Init GPIO for encoders outputs
 *
 */
void initEncoderOutputs(void) { 
	
	/*
	 * Setting quadrature encoder outputs:
	 *
	 * Input pins:
	 * 	- MOTOR_A:
	 * 		OUTA --> PD1
	 * 		OUTB --> not set;
	 *
	 * 	- MOTOR_B:
	 * 		OUTA --> PD0
	 * 		OUTB --> not set;
	 */
	 
	/* Setting PD0 and PD1 as input pins. */
	DDRD &= ~(1 << PD0); /* MOTOR_B - OUTB */
	DDRD &= ~(1 << PD1); /* MOTOR_A - OUTB */
	
	/* Activez rezistenta de pull-up pentru pinii PD0, PD1. */
	PORTD |= (1 << PD0);
	PORTD |= (1 << PD1);
	
}

/*
 * Init GPIO for motors direction.
 *
 */
void initDirection(void) {
	
	/*
	 * Setting motors direction:
	 *
	 * Output pins:
	 * 		AIN1 --> PB6, AIN2 --> PB1
	 * 		BIN1 --> PB5, BIN2 --> PB0
	 */
	
	/* Setting PB0, PB1, PB5, PB6 as output pins. */
	DDRB |= (1 << PB0); //BIN2
	DDRB |= (1 << PB1); //AIN2
	DDRB |= (1 << PB5); //BIN1
	DDRB |= (1 << PB6); //AIN1
	
	/* Put the motors in FLOATing state. */
	PORTB |= (0 << PB0);
	PORTB |= (0 << PB1);
	PORTB |= (0 << PB5);
	PORTB |= (0 << PB6);
}



/*
 * Init GPIO for PWM
 *
 * - init GPIO PD6, PD7
 * - timer2 is used
 *
 */
 
void initPWM(void)
{
	/* Setting PD6 as output pins (OC2B). */
	DDRD |= (1 << PD6);
	/* Setting PD7 as output pins (OC2A). */
	DDRD |= (1 << PD7); 
		
	/* Set fast pwm 8-bit, top at 0xFF */
	/* non-inverting mode */
	TCCR2A |= (1 << WGM21) | (1 << WGM20) | (1 << COM2A1) | (1 << COM2B1);
	
	/* set Prescaler - 1024 (CS02 = 1 and CS00 = 1) */
	TCCR2B |= (1 << CS20);
}



/*
 * Set motor speed for the given motor
 *
 * motor 	- motor_A or motor_B
 * speed  	- value of the duty cycle
 *
 */
void setPwm(Motor motor, int speed){
	if(motor == MOTOR_A) {
		OCR2A = speed;
	
	} else if(motor == MOTOR_B) {
		OCR2B = speed;
	}
}





/*
 * Function to set the direction of the given motor
 *
 * motor 	- motor_A or motor_B
 * dir 		- Float/FORWARD/BACKWARD/BREAK
 *
 * ###	AIN1 --> PB0, AIN2 --> PB5
 * ###	BIN1 --> PB1, BIN2 --> PB6
 */
void setDirection(Motor motor, Direction dir) {

	if(motor == MOTOR_A) {

		if(dir == FLOAT) { //00
			PORTB |= (0 << PB0) | (0 << PB5);
			
		
		} else if (dir == FORWARD) { //01
			PORTB |= (0 << PB0) | (1 << PB5);
		
		
		} else if (dir == BACKWARD) { //10
			PORTB |= (1 << PB0) | (0 << PB5);
		
		
		} else if (dir == BREAK) { //11
			PORTB |= (1 << PB0) | (1 << PB5);
			
		
		}
	
	
	} else if(motor == MOTOR_B) {

		if(dir == FLOAT) {
			PORTB |= (0 << PB1) | (0 << PB6);
	
		
		} else if (dir == FORWARD) {
			PORTB |= (0 << PB1) | (1 << PB6);
		
		
		} else if (dir == BACKWARD) {
			PORTB |= (1 << PB1) | (0 << PB6);
		
		
		} else if (dir == BREAK) {
			PORTB |= (1 << PB1) | (1 << PB6);
		
		}
	}
}


/*
 * Move the wheels in the direction given by the sign of velocity.
 * 		va - right wheel velocity
 * 		vb - left wheel velocity

 */
void move(int va, int vb) {
	Direction dir;
	
	if (va < 0) 
		dir = BACKWARD;
		
	else if (va > 0)
		dir = FORWARD;
	
	else 
		dir = FLOAT;
		
	setDirection(MOTOR_A, dir);
	
	setPwm(MOTOR_A, va);
	
	
	if (vb < 0) 
		dir = BACKWARD;
	
	else if (va > 0)
		dir = FORWARD;
	
	else 
		dir = FLOAT;
		
	setDirection(MOTOR_B, dir);
	
	setPwm(MOTOR_B, vb);
}



/* Set timer timer to trigger update once at 625 counts (every 10ms). */
ISR(TIMER2_OVF_vect)
{
   PWM1 ++;
   
   if (PWM1 == 156 && PWM2 != 4) {
	
		PWM1 = 0;
		PWM2 ++;
		
		if( PWM2 != 4)
			update = 1;
	
   } else if (PWM2 == 4) {
		//test - aprind LED
		PORTD |= (1 << PD7);
		
		PWM1 = 0;
		PWM2 = 0;
		
		update = 1;
   }
}



/*
 * Init Timer0 to count using the clock given by motorA, OUTB.
 * top at 0xFF
 * EXTCLK must be set
 */
void timer0_init()
{
    TCCR0A = (3 << WGM00) | (2 << COM0A0);
    TCCR0B = (1 << CS00) | (1 << CS01) | (1 << CS02);
    OCR0A = 0xFF;//numara pana la oxff
}



/*
 * Init Timer1 to count using the clock given by motorB, OUTB.
 * top at 0xFF
 * EXTCLK must be set
 */
void timer1_init()
{
    TCCR1A = (3 << WGM10) | (2 << COM1A0);
    TCCR1B = (1 << CS10) | (1 << CS11) | (1 << CS12);
    OCR1A = 0xFF;
}

/*
 * Init GPIO for encoders counters
 *
 */
void initEcoderCounters(void)
{
	/* Timer0 */
	timer0_init();
	
	/* Timer1 */
	timer1_init();
}

/*
 * Read the value of counter, treating overflow incontinence.
 */
short readCounter(Counter *c, Direction dir, Motor m) {

	short sum = 0;
	
	c->old = c->curr;
	
	if (m == MOTOR_A) /* right */
		c->curr = TCNT1;
	else if (m == MOTOR_B) /* left */
		c->curr = TCNT2;
	
	
	//char s = (dir == 0) ? -1 : 1;
	char s = dir; /* sign */

	if (m == MOTOR_A) { /* right */
		sumr += s * (c->curr - c->old);
		sum = sumr;
		
	} else if (m == MOTOR_B) { /* left */
		suml += s * (c->curr - c->old);
		sum = suml;
	}
	

	return sum;
}

/*
 * Convert the counts in distance.
 */
short readEncoder(Encoder *e, Direction dir, Motor m)
{
	e->old = e->curr;
	
	if(m == MOTOR_A) /* right */
		e->curr = readCounter(cr, dir, m);
	
	
	else if(m == MOTOR_B) /* left */
		e->curr = readCounter(cl, dir, m);
	
	
	short delta = e->curr - e->old;
	
	
	if (delta < 0 && dir == FORWARD)	
		delta += ENC_OVF;
		
	else if (delta < 0 && dir == BACKWARD)
		delta -= ENC_OVF;
	
	
	return delta;
}


//actualizez pozitia robotului pe baza outputului de la encodere
/*
 * Update robot position using encoders output.
 */
void updateRobotPosition(void){

	/* read form encoder the distance */
	long dl = 2 * pi * R * readEncoder(el, FORWARD, MOTOR_B) / N;
	long dr = 2 * pi * R * readEncoder(er, FORWARD, MOTOR_A) / N;
	long dc = (dl + dr)/2;

	/* update robot position */
	robotPosition->x += dc * cos(robotPosition->teta);
	robotPosition->y += dc * sin(robotPosition->teta);
	robotPosition->teta = robotPosition->teta + (dl-dr)/2;
	robotPosition->teta = atan2 (sin(robotPosition->teta), cos(robotPosition->teta)); /* normalize teta */
}

/* 
 * Set the robot directions and speed to performe a rotation.
 */
int rotate(long tetaGoal, int speed){

	//calculez unghiul de rotatie
	long deltaTeta = robotPosition->teta - atan2(sin(tetaGoal), cos(tetaGoal));

	/* if deltaTeta is less than 10 radians, we don't mve */
	if( abs(deltaTeta - ANGLE_ERROR_THRESHOLD) <= 0){
		move(0, 0);
		return 0;
	
	/* else we are moving */
	} else {
		move( sign(deltaTeta) * speed, -sign(deltaTeta) * speed );   
		return 1;
	}
}

/*
 * Function to travel the distance.
 */
int travelDistance(long xGoal, long yGoal, int speed){ 


	long deltaDistance = 
		sqrt( (xGoal - robotPosition->x) * (xGoal - robotPosition->x) +
			  (yGoal - robotPosition->y) * (yGoal - robotPosition->y));
			  
	if (deltaDistance > 0)
		move(speed, speed );
	else 
		move(0, 0);
		
	return 1;
}




int main(void) {

	robotPosition = (Position *)malloc (sizeof(Position));
	cl = (Counter *)malloc(sizeof(Counter));
	cr = (Counter *)malloc(sizeof(Counter));
	el = (Encoder *)malloc(sizeof(Encoder));
	er = (Encoder *)malloc(sizeof(Encoder));
	
	/*	init GPIO for motor directions */
	initDirection();
	
	/* init GPIO for encoders outputs */
	initEncoderOutputs();
	
	/* init PWM (PWMA, PWMB) for both motors */
	initPWM();

	/* init timer0 si timer1 for encoders counters */
	initEcoderCounters(); 

	sei();
	
		
	/* Initial position of the robot. */
	robotPosition->x = 0;
	robotPosition->y = 0;
	robotPosition->teta = 0;
	
	/* Set the final goal */
	goal.x = 3;
	goal.y = 6;
	
	/* flags for moving */
	int rotationComplete = 0;
	int travelComplete = 0;
	
	int speed = 50;

	setDirection(MOTOR_A, FORWARD);
	setDirection(MOTOR_B, FORWARD);
	
	while(1){

		if(update){
			updateRobotPosition();

			if(!rotationComplete)
				rotationComplete = rotate(goal.teta, speed);
			 
			if(rotationComplete && !travelComplete)
				travelDistance(goal.x, goal.y, speed);

			update = 0;
		}

	}
	
	return 0;
}



